/*
 * calibrate_RGB.cpp
 *
 *
 * Created on: Feb 15,2016
 * 		Author: Arantxa Casanova
 */


#include <imageplus/kinect_calibration/calibrate_RGB.hpp>

#define PI 3.1416

using namespace cv;
using namespace std;
using namespace imageplus;
using namespace imageplus::kinect_calibration;
using namespace pcl;

/*!@file
 */

void Settings::write(FileStorage& fs) const //Write serialization for this class
		{
	//!Write values to output XML file.
	/*!
	 \sa write()
	 \param fs XML output file.
	 */
	fs << "{" << "BoardSize_in_Width" << boardSize_in.width
			<< "BoardSize_in_Height" << boardSize_in.height << "Square_Size_in"
			<< squareSize_in << "BoardSize_ex_Width" << boardSize_ex.width
			<< "BoardSize_ex_Height" << boardSize_ex.height << "Square_Size_ex"
			<< squareSize_ex << "Calibrate_Pattern" << patternToUse
			<< "Calibrate_NrOfFrameToUse" << nrFrames
			<< "Calibrate_FixAspectRatio" << aspectRatio
			<< "Calibrate_AssumeZeroTangentialDistortion"
			<< calibZeroTangentDist << "Calibrate_FixPrincipalPointAtTheCenter"
			<< calibFixPrincipalPoint

			<< "Write_DetectedFeaturePoints" << bwritePoints
			<< "Write_extrinsicParameters" << bwriteExtrinsics
			<< "Write_in_outputFileName" << outputFileName_in
			<< "Write_ex_outputFileName" << outputFileName_ex

			<< "Show_UndistortedImage" << showUndistorsed

			<< "Input_FlipAroundHorizontalAxis" << flipVertical << "Input_Delay"
			<< delay << "Input_in" << input_in << "Input_ex" << input_ex << "}";
}
void Settings::read(const FileNode& node)    //Read serialization for this class
		{
	node["BoardSize_in_Width"] >> boardSize_in.width;
	node["BoardSize_in_Height"] >> boardSize_in.height;
	node["BoardSize_ex_Width"] >> boardSize_ex.width;
	node["BoardSize_ex_Height"] >> boardSize_ex.height;
	node["Calibrate_Pattern"] >> patternToUse;
	node["Square_Size_in"] >> squareSize_in;
	node["Square_Size_ex"] >> squareSize_ex;
	node["Calibrate_NrOfFrameToUse"] >> nrFrames;
	node["Calibrate_FixAspectRatio"] >> aspectRatio;
	node["Write_DetectedFeaturePoints"] >> bwritePoints;
	node["Write_extrinsicParameters"] >> bwriteExtrinsics;
	node["Write_in_outputFileName"] >> outputFileName_in;
	node["Write_ex_outputFileName"] >> outputFileName_ex;
	node["Calibrate_AssumeZeroTangentialDistortion"] >> calibZeroTangentDist;
	node["Calibrate_FixPrincipalPointAtTheCenter"] >> calibFixPrincipalPoint;
	node["Input_FlipAroundHorizontalAxis"] >> flipVertical;
	node["Show_UndistortedImage"] >> showUndistorsed;
	node["Input_in"] >> input_in;
	node["Input_ex"] >> input_ex;
	node["Input_Delay"] >> delay;
	node["CornersFileName_ex"] >> cornersFileName_ex;
	interprate();
}
void Settings::interprate() {
	goodInput = true;
	if (boardSize_in.width <= 0 || boardSize_in.height <= 0) {
		cerr << "Invalid Board size: " << boardSize_in.width << " "
				<< boardSize_in.height << endl;
		goodInput = false;
	}
	if (squareSize_in <= 10e-6) {
		cerr << "Invalid square size " << squareSize_in << endl;
		goodInput = false;
	}
	if (nrFrames <= 0) {
		cerr << "Invalid number of frames " << nrFrames << endl;
		goodInput = false;
	}

	if (input_in.empty() || input_ex.empty())      // Check for valid input
		inputType = INVALID;
	else {
		if (input_in[0] >= '0' && input_in[0] <= '9') {
			stringstream ss(input_in);
			ss >> cameraID;
			inputType = CAMERA;
		} else {
			if (readStringList(input_in, imageList_in)) {
				inputType = IMAGE_LIST;
				nrFrames =
						(nrFrames < (int) imageList_in.size()) ?
								nrFrames : (int) imageList_in.size();
			} else
				inputType = VIDEO_FILE;

			readStringList(input_ex, imageList_ex);

		}
		if (inputType == CAMERA)
			inputCapture.open(cameraID);
		if (inputType == VIDEO_FILE)
			inputCapture.open(input_in);
		if (inputType != IMAGE_LIST && !inputCapture.isOpened())
			inputType = INVALID;
	}
	if (inputType == INVALID) {
		cerr << " Inexistent input: " << input_in << "or inexistent input: "
				<< input_ex;
		goodInput = false;
	}

	flag = 0;
	if (calibFixPrincipalPoint)
		flag |= CV_CALIB_FIX_PRINCIPAL_POINT;
	if (calibZeroTangentDist)
		flag |= CV_CALIB_ZERO_TANGENT_DIST;
	if (aspectRatio)
		flag |= CV_CALIB_FIX_ASPECT_RATIO;

	calibrationPattern = NOT_EXISTING;
	if (!patternToUse.compare("CHESSBOARD"))
		calibrationPattern = CHESSBOARD;
	if (!patternToUse.compare("CIRCLES_GRID"))
		calibrationPattern = CIRCLES_GRID;
	if (!patternToUse.compare("ASYMMETRIC_CIRCLES_GRID"))
		calibrationPattern = ASYMMETRIC_CIRCLES_GRID;
	if (calibrationPattern == NOT_EXISTING) {
		cerr << " Inexistent camera calibration mode: " << patternToUse << endl;
		goodInput = false;
	}
	atImageList_in = 0;
	atImageList_ex = 0;

}
Mat Settings::nextImage(int& atImageList, vector<string> imageList) {
	Mat result;
	if (inputCapture.isOpened()) {
		Mat view0;
		inputCapture >> view0;
		view0.copyTo(result);
	} else if (atImageList < (int) imageList.size()) {
		result = imread(imageList[atImageList++], CV_LOAD_IMAGE_COLOR);
	}

	return result;
}

bool Settings::readStringList(const string& filename, vector<string>& l) {
	l.clear();
	FileStorage fs(filename, FileStorage::READ);
	if (!fs.isOpened())
		return false;
	FileNode n = fs.getFirstTopLevelNode();
	if (n.type() != FileNode::SEQ)
		return false;
	FileNodeIterator it = n.begin(), it_end = n.end();
	for (; it != it_end; ++it)
		l.push_back((string) * it);
	return true;
}

void read(const FileNode& node, Settings& x, const Settings& default_value =
		Settings()) {
	if (node.empty())
		x = default_value;
	else
		x.read(node);
}

enum {
	DETECTION = 0, CAPTURING = 1, CALIBRATED = 2
};

int imageplus::kinect_calibration::get_intrinsics(Settings& s) {

	vector < vector<Point2f> > imagePoints;
	Mat cameraMatrix, distCoeffs;
	Size imageSize;
	clock_t prevTimestamp = 0;
	int mode = s.inputType == Settings::IMAGE_LIST ? CAPTURING : DETECTION;
	const Scalar RED(0, 0, 255), GREEN(0, 255, 0);
	const char ESC_KEY = 27;

	for (int i = 0;; ++i) {
		Mat view;
		bool blinkOutput = false;

		view = s.nextImage(s.atImageList_in, s.imageList_in);

		if (view.empty()) // If no more images then run calibration, save and stop loop.
		{
			if (imagePoints.size() > 0) {

				runCalibrationAndSave(s, imageSize, cameraMatrix, distCoeffs,
						imagePoints);

			}
			break;
		}

		imageSize = view.size();  // Format input image.
		cout << imageSize << endl;
		if (s.flipVertical)
			flip(view, view, 0);

		vector < Point2f > pointBuf;

		bool found;
		found = findChessboardCorners(view, s.boardSize_in, pointBuf,
				CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_FAST_CHECK
						| CV_CALIB_CB_NORMALIZE_IMAGE);

		if (found)                // If done with success,
		{
			// improve the found corners' coordinate accuracy for chessboard
			cout << "found!!" << endl;
			Mat viewGray;
			cvtColor(view, viewGray, COLOR_BGR2GRAY);
			cornerSubPix(viewGray, pointBuf, Size(11, 11), Size(-1, -1),
					TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));

			if (mode == CAPTURING && // For camera only take new samples after delay time
					(!s.inputCapture.isOpened()
							|| clock() - prevTimestamp
									> s.delay * 1e-3 * CLOCKS_PER_SEC)) {
				imagePoints.push_back(pointBuf);
				prevTimestamp = clock();
				blinkOutput = s.inputCapture.isOpened();
			}

			// Draw the corners.
			drawChessboardCorners(view, s.boardSize_in, Mat(pointBuf), found);
		}

		//----------------------------- Output Text ------------------------------------------------
		string msg =
				(mode == CAPTURING) ? "100/100" :
				mode == CALIBRATED ? "Calibrated" : "Press 'g' to start";
		int baseLine = 0;
		Size textSize = getTextSize(msg, 1, 1, 1, &baseLine);
		Point textOrigin(view.cols - 2 * textSize.width - 10,
				view.rows - 2 * baseLine - 10);

		if (mode == CAPTURING) {
			if (s.showUndistorsed)
				msg = format("%d/%d Undist", (int) imagePoints.size(),
						s.nrFrames);
			else
				msg = format("%d/%d", (int) imagePoints.size(), s.nrFrames);
		}

		putText(view, msg, textOrigin, 1, 1, mode == CALIBRATED ? GREEN : RED);

		if (blinkOutput)
			bitwise_not(view, view);

		//------------------------- Video capture  output  undistorted ------------------------------
		if (mode == CALIBRATED && s.showUndistorsed) {
			Mat temp = view.clone();
			undistort(temp, view, cameraMatrix, distCoeffs);
		}

		//------------------------------ Show image and check for input commands -------------------
		imshow("Image View", view);
		char key = (char) waitKey(s.inputCapture.isOpened() ? 50 : s.delay);

		if (key == ESC_KEY)
			break;

		if (key == 'u' && mode == CALIBRATED)
			s.showUndistorsed = !s.showUndistorsed;

		if (s.inputCapture.isOpened() && key == 'g') {
			mode = CAPTURING;
			imagePoints.clear();
		}
	}

	// -----------------------Show the undistorted image for the image list ------------------------
	if (s.inputType == Settings::IMAGE_LIST && s.showUndistorsed) {
		Mat view, rview, map1, map2;
		initUndistortRectifyMap(cameraMatrix, distCoeffs, Mat(),
				getOptimalNewCameraMatrix(cameraMatrix, distCoeffs, imageSize,
						1, imageSize, 0), imageSize, CV_16SC2, map1, map2);

		for (int i = 0; i < (int) s.imageList_in.size(); i++) {
			view = imread(s.imageList_in[i], 1);
			if (view.empty())
				continue;
			remap(view, rview, map1, map2, INTER_LINEAR);
			imshow("Image View", rview);
			char c = (char) waitKey();
			if (c == ESC_KEY || c == 'q' || c == 'Q')
				break;
		}
	}
	return 0;
}

int imageplus::kinect_calibration::get_extrinsics(Settings& s) {

	Mat cameraMatrix, distCoeffs;

	FileStorage fs2(s.outputFileName_in, FileStorage::READ); // Read the settings
	if (!fs2.isOpened()) {
		cout << "Could not open the output intrinsics file: \""
				<< s.outputFileName_in << "\"" << endl;
		return -1;
	}

	fs2["Camera_Matrix"] >> cameraMatrix;
	fs2["Distortion_Coefficients"] >> distCoeffs;
	fs2.release();

	vector < vector<Point2f> > imagePoints;
	vector < vector<Point3f> > imagePoints_h;

	Size imageSize;
	clock_t prevTimestamp = 0;
	int mode = s.inputType == Settings::IMAGE_LIST ? CAPTURING : DETECTION;
	const Scalar RED(0, 0, 255), GREEN(0, 255, 0);
	const char ESC_KEY = 27;

	for (int i = 0;; ++i) {
		Mat ycrcb, view;
		bool blinkOutput = false;

		view = s.nextImage(s.atImageList_ex, s.imageList_ex);

		/*  cvtColor(view,ycrcb,CV_BGR2YCrCb);

		 vector<Mat> channels;
		 split(ycrcb,channels);

		 equalizeHist(channels[0], channels[0]);

		 Mat result,temp;
		 merge(channels,ycrcb);

		 cvtColor(ycrcb,view,CV_YCrCb2BGR);

		 GaussianBlur(view,temp, Size(0,0), 105) ; //hardcoded filter size, to be tested on 50 mm lens
		 addWeighted(view, 1.8, temp, -0.8,0,view) ; //hardcoded weight, to be tested.
		 */
		if (imagePoints.size() > 0) {
			runCalibrationExtrinsicsAndSave(s, imageSize, cameraMatrix,
					distCoeffs, imagePoints);
			break;
		}

		imageSize = view.size();  // Format input image.
		if (s.flipVertical)
			flip(view, view, 0);

		vector < Point2f > pointBuf;

		bool found;
		cout << "abans find chess" << endl;
		found = findChessboardCorners(view, s.boardSize_ex, pointBuf,
				CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_FAST_CHECK
						| CV_CALIB_CB_NORMALIZE_IMAGE);

		if (found)                // If done with success,
		{
			cout << "found!!" << endl;
			// improve the found corners' coordinate accuracy for chessboard

			Mat viewGray;
			cvtColor(view, viewGray, COLOR_BGR2GRAY);
			cornerSubPix(viewGray, pointBuf, Size(2, 2), Size(-1, -1),
					TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));

			if (mode == CAPTURING && // For camera only take new samples after delay time
					(!s.inputCapture.isOpened()
							|| clock() - prevTimestamp
									> s.delay * 1e-3 * CLOCKS_PER_SEC)) {
				imagePoints.push_back(pointBuf);
				prevTimestamp = clock();
				blinkOutput = s.inputCapture.isOpened();
			}

			// Draw the corners.
			drawChessboardCorners(view, s.boardSize_ex, Mat(pointBuf), found);

		}

		//----------------------------- Output Text ------------------------------------------------
		string msg =
				(mode == CAPTURING) ? "100/100" :
				mode == CALIBRATED ? "Calibrated" : "Press 'g' to start";
		int baseLine = 0;
		Size textSize = getTextSize(msg, 1, 1, 1, &baseLine);
		Point textOrigin(view.cols - 2 * textSize.width - 10,
				view.rows - 2 * baseLine - 10);

		if (mode == CAPTURING) {
			if (s.showUndistorsed)
				msg = format("%d/%d Undist", (int) imagePoints.size(),
						s.nrFrames);
			else
				msg = format("%d/%d", (int) imagePoints.size(), s.nrFrames);
		}

		putText(view, msg, textOrigin, 1, 1, mode == CALIBRATED ? GREEN : RED);

		if (blinkOutput)
			bitwise_not(view, view);

		//------------------------- Video capture  output  undistorted ------------------------------
		if (mode == CALIBRATED && s.showUndistorsed) {
			Mat temp = view.clone();
			undistort(temp, view, cameraMatrix, distCoeffs);
		}

		//------------------------------ Show image and check for input commands -------------------
		imshow("Image View", view);
		char key = (char) waitKey();

		if (key == ESC_KEY)
			break;

		if (key == 'u' && mode == CALIBRATED)
			s.showUndistorsed = !s.showUndistorsed;

		if (s.inputCapture.isOpened() && key == 'g') {
			mode = CAPTURING;
			imagePoints.clear();
		}
	}

	// -----------------------Show the undistorted image for the image list ------------------------
	if (s.inputType == Settings::IMAGE_LIST && s.showUndistorsed) {
		Mat view, rview, map1, map2;
		initUndistortRectifyMap(cameraMatrix, distCoeffs, Mat(),
				getOptimalNewCameraMatrix(cameraMatrix, distCoeffs, imageSize,
						1, imageSize, 0), imageSize, CV_16SC2, map1, map2);

		for (int i = 0; i < (int) s.imageList_ex.size(); i++) {
			view = imread(s.imageList_ex[i], 1);
			if (view.empty())
				continue;
			remap(view, rview, map1, map2, INTER_LINEAR);
			imshow("Image View", rview);
			char c = (char) waitKey();

			if (c == ESC_KEY || c == 'q' || c == 'Q')
				break;
		}
	}
	return 0;
}

void kinect_calibration::reproject_2Dto3D() {

	Mat cameraMatrix, distCoeffs, rvecs, tvecs, aux_mul, t_point, corners_3d;
	vector<Point2f> calc_corners, und_corners;
	vector<Point3f> calc_corners_h, und_corners_h;

	FileStorage out_in("out_in_asus.xml", FileStorage::READ);
	if (!out_in.isOpened()) {
		cout << "Could not open the configuration file: \\" << endl;

	}
	out_in["Camera_Matrix"] >> cameraMatrix;

	out_in["Distortion_Coefficients"] >> distCoeffs;
	out_in.release();

	//get extrinsics matrix (rotation and translation vectors)

	FileStorage out_ex("out_ex_asus.xml", FileStorage::READ);
	if (!out_ex.isOpened()) {
		cout << "Could not open the configuration file: \\" << endl;

	}
	out_ex["rvecs"] >> rvecs;
	out_ex["tvecs"] >> tvecs;
	out_ex["calculated_corners"] >> calc_corners;
	out_ex.release();
	cout << calc_corners << endl;
	//re-project detected corners into 3D world
	convertPointsToHomogeneous(calc_corners, calc_corners_h);
	cout << calc_corners_h << endl;

	Mat rotation_matrix, reprojection_matrix, reproj_corners;
	reprojection_matrix = Mat::zeros(3, 2, CV_64F);
	t_point = Mat::zeros(3, 1, CV_64FC1);
	reproj_corners = Mat::zeros(3, 3, CV_64FC1);
	corners_3d = Mat::zeros(4, 1, CV_64FC1);

	Rodrigues(rvecs, rotation_matrix);
	hconcat(rotation_matrix, tvecs, rotation_matrix);

	undistortPoints(calc_corners, und_corners, cameraMatrix, distCoeffs);
	convertPointsToHomogeneous(und_corners, und_corners_h);

}

int imageplus::kinect_calibration::getCalibrationParameters(
		std::string config_file) {
	cout << "Option: calculate extrinsics" << endl;
	Settings s;
	FileStorage fs(config_file, FileStorage::READ); // Read the settings

	if (!fs.isOpened()) {
		cout << "Could not open the configuration file: \"" << config_file
				<< "\"" << endl;
		return -1;
	}
	fs["Settings"] >> s;
	fs.release();                                         // close Settings file

	if (!s.goodInput) {
		cout << "Invalid input detected. Application stopping. " << endl;
		return -1;
	}

	//
	//  get_intrinsics(s);
	// cout << "Intrinsics calculated"<< endl;
	get_extrinsics(s);

	return 0;
}

double imageplus::kinect_calibration::computeReprojectionErrors(
		const vector<vector<Point3f> >& objectPoints,
		const vector<vector<Point2f> >& imagePoints, const vector<Mat>& rvecs,
		const vector<Mat>& tvecs, const Mat& cameraMatrix,
		const Mat& distCoeffs, vector<float>& perViewErrors) {
	vector < Point2f > imagePoints2;
	int i, totalPoints = 0;
	double totalErr = 0, err;
	perViewErrors.resize(objectPoints.size());

	for (i = 0; i < (int) objectPoints.size(); ++i) {
		projectPoints(Mat(objectPoints[i]), rvecs[i], tvecs[i], cameraMatrix,
				distCoeffs, imagePoints2);
		err = norm(Mat(imagePoints[i]), Mat(imagePoints2), CV_L2);

		int n = (int) objectPoints[i].size();
		perViewErrors[i] = (float) std::sqrt(err * err / n);
		totalErr += err * err;
		totalPoints += n;
	}

	return std::sqrt(totalErr / totalPoints);
}

void imageplus::kinect_calibration::calcBoardCornerPositions(Size boardSize,
		float squareSize, vector<Point3f>& corners,
		Settings::Pattern patternType = Settings::CHESSBOARD) {
	corners.clear();

	switch (patternType) {
	case Settings::CHESSBOARD:
	case Settings::CIRCLES_GRID:
		for (int i = 0; i < boardSize.height; ++i)
			for (int j = 0; j < boardSize.width; ++j)
				corners.push_back(
						Point3f(float(j * squareSize), float(i * squareSize),
								0));
		break;

	case Settings::ASYMMETRIC_CIRCLES_GRID:
		for (int i = 0; i < boardSize.height; i++)
			for (int j = 0; j < boardSize.width; j++)
				corners.push_back(
						Point3f(float((2 * j + i % 2) * squareSize),
								float(i * squareSize), 0));
		break;
	default:
		break;
	}
}

bool imageplus::kinect_calibration::runCalibration(Settings& s, Size& imageSize,
		Mat& cameraMatrix, Mat& distCoeffs,
		vector<vector<Point2f> > imagePoints, vector<Mat>& rvecs,
		vector<Mat>& tvecs, vector<float>& reprojErrs, double& totalAvgErr) {

	cameraMatrix = Mat::eye(3, 3, CV_64F);
	if (s.flag & CV_CALIB_FIX_ASPECT_RATIO)
		cameraMatrix.at<double>(0, 0) = 1.0;

	distCoeffs = Mat::zeros(8, 1, CV_64F);

	vector < vector<Point3f> > objectPoints(1);
	calcBoardCornerPositions(s.boardSize_in, s.squareSize_in, objectPoints[0],
			s.calibrationPattern);

	objectPoints.resize(imagePoints.size(), objectPoints[0]);

	//Find intrinsic and extrinsic camera parameters
	double rms = calibrateCamera(objectPoints, imagePoints, imageSize,
			cameraMatrix, distCoeffs, rvecs, tvecs,
			s.flag | CV_CALIB_FIX_K4 | CV_CALIB_FIX_K5);

	cout << "Re-projection error reported by calibrateCamera: " << rms << endl;

	bool ok = checkRange(cameraMatrix) && checkRange(distCoeffs);

	totalAvgErr = computeReprojectionErrors(objectPoints, imagePoints, rvecs,
			tvecs, cameraMatrix, distCoeffs, reprojErrs);

	return ok;
}

bool imageplus::kinect_calibration::runCalibration_extrinsics(Settings& s,
		Size& imageSize, Mat cameraMatrix, Mat distCoeffs,
		vector<vector<Point2f> > imagePoints, Mat& rvecs, Mat& tvecs,
		vector<float>& reprojErrs, double& totalAvgErr) {


	vector < vector<Point3f> > objectPoints(1);
	calcBoardCornerPositions(s.boardSize_ex, s.squareSize_ex, objectPoints[0],
			s.calibrationPattern);

	objectPoints.resize(imagePoints.size(), objectPoints[0]);

	//extrinsic camera parameters

	solvePnP(objectPoints[0], imagePoints[0], cameraMatrix, distCoeffs, rvecs,
			tvecs);
	bool ok = checkRange(cameraMatrix) && checkRange(distCoeffs);

	return ok;
}

// Print camera parameters to the output file
void imageplus::kinect_calibration::saveCameraParams(Settings& s,
		Size& imageSize, Mat& cameraMatrix, Mat& distCoeffs,
		const vector<Mat>& rvecs, const vector<Mat>& tvecs,
		const vector<float>& reprojErrs,
		const vector<vector<Point2f> >& imagePoints, double totalAvgErr) {
	FileStorage fs(s.outputFileName_in, FileStorage::WRITE);

	time_t tm;
	time(&tm);
	struct tm *t2 = localtime(&tm);
	char buf[1024];
	strftime(buf, sizeof(buf) - 1, "%c", t2);

	fs << "calibration_Time" << buf;

	if (!rvecs.empty() || !reprojErrs.empty())
		fs << "nrOfFrames" << (int) std::max(rvecs.size(), reprojErrs.size());
	fs << "image_Width" << imageSize.width;
	fs << "image_Height" << imageSize.height;
	fs << "board_Width_in" << s.boardSize_in.width;
	fs << "board_Height_in" << s.boardSize_in.height;
	fs << "square_Size_in" << s.squareSize_in;

	if (s.flag & CV_CALIB_FIX_ASPECT_RATIO)
		fs << "FixAspectRatio" << s.aspectRatio;

	if (s.flag) {
		sprintf(buf, "flags: %s%s%s%s",
				s.flag & CV_CALIB_USE_INTRINSIC_GUESS ?
						" +use_intrinsic_guess" : "",
				s.flag & CV_CALIB_FIX_ASPECT_RATIO ? " +fix_aspectRatio" : "",
				s.flag & CV_CALIB_FIX_PRINCIPAL_POINT ?
						" +fix_principal_point" : "",
				s.flag & CV_CALIB_ZERO_TANGENT_DIST ?
						" +zero_tangent_dist" : "");
		cvWriteComment(*fs, buf, 0);

	}

	fs << "flagValue" << s.flag;

	fs << "Camera_Matrix" << cameraMatrix;
	fs << "Distortion_Coefficients" << distCoeffs;

	fs << "Avg_Reprojection_Error" << totalAvgErr;
	if (!reprojErrs.empty())
		fs << "Per_View_Reprojection_Errors" << Mat(reprojErrs);

	if (!rvecs.empty() && !tvecs.empty()) {
		CV_Assert(rvecs[0].type() == tvecs[0].type());
		Mat bigmat((int) rvecs.size(), 6, rvecs[0].type());
		for (int i = 0; i < (int) rvecs.size(); i++) {
			Mat r = bigmat(Range(i, i + 1), Range(0, 3));
			Mat t = bigmat(Range(i, i + 1), Range(3, 6));

			CV_Assert(rvecs[i].rows == 3 && rvecs[i].cols == 1);
			CV_Assert(tvecs[i].rows == 3 && tvecs[i].cols == 1);
			//*.t() is MatExpr (not Mat) so we can use assignment operator
			r = rvecs[i].t();
			t = tvecs[i].t();
		}
		cvWriteComment(*fs,
				"a set of 6-tuples (rotation vector + translation vector) for each view",
				0);
		fs << "Extrinsic_Parameters" << bigmat;
	}

	if (!imagePoints.empty()) {
		Mat imagePtMat((int) imagePoints.size(), (int) imagePoints[0].size(),
				CV_32FC2);
		for (int i = 0; i < (int) imagePoints.size(); i++) {
			Mat r = imagePtMat.row(i).reshape(2, imagePtMat.cols);
			Mat imgpti(imagePoints[i]);
			imgpti.copyTo(r);
		}
		fs << "Image_points" << imagePtMat;
	}
}

void imageplus::kinect_calibration::saveCameraParams_extrinsics(Settings& s,
		Size& imageSize, Mat& cameraMatrix, Mat& distCoeffs, const Mat& rvecs,
		const Mat& tvecs, const vector<float>& reprojErrs,
		const vector<vector<Point2f> >& imagePoints, double totalAvgErr) {
	FileStorage fs(s.outputFileName_ex, FileStorage::WRITE);

	time_t tm;
	time(&tm);
	struct tm *t2 = localtime(&tm);
	char buf[1024];
	strftime(buf, sizeof(buf) - 1, "%c", t2);

	fs << "calibration_Time" << buf;

	if (!rvecs.empty() || !reprojErrs.empty())
		fs << "image_Width" << imageSize.width;
	fs << "image_Height" << imageSize.height;
	fs << "board_Width_ex" << s.boardSize_ex.width;
	fs << "board_Height_ex" << s.boardSize_ex.height;
	fs << "square_Size_ex" << s.squareSize_ex;

	if (s.flag & CV_CALIB_FIX_ASPECT_RATIO)
		fs << "FixAspectRatio" << s.aspectRatio;

	if (s.flag) {
		sprintf(buf, "flags: %s%s%s%s",
				s.flag & CV_CALIB_USE_INTRINSIC_GUESS ?
						" +use_intrinsic_guess" : "",
				s.flag & CV_CALIB_FIX_ASPECT_RATIO ? " +fix_aspectRatio" : "",
				s.flag & CV_CALIB_FIX_PRINCIPAL_POINT ?
						" +fix_principal_point" : "",
				s.flag & CV_CALIB_ZERO_TANGENT_DIST ?
						" +zero_tangent_dist" : "");
		cvWriteComment(*fs, buf, 0);

	}

	fs << "flagValue" << s.flag;

	fs << "Camera_Matrix" << cameraMatrix;
	fs << "Distortion_Coefficients" << distCoeffs;

	fs << "Avg_Reprojection_Error" << totalAvgErr;
	fs << "calculated_corners" << imagePoints[0];
	if (!reprojErrs.empty())
		fs << "Per_View_Reprojection_Errors" << Mat(reprojErrs);

	if (!rvecs.empty() && !tvecs.empty()) {

		fs << "rvecs" << rvecs;

		fs << "tvecs" << tvecs;
	}

	if (!imagePoints.empty()) {
		Mat imagePtMat((int) imagePoints.size(), (int) imagePoints[0].size(),
				CV_32FC2);
		for (int i = 0; i < (int) imagePoints.size(); i++) {
			Mat r = imagePtMat.row(i).reshape(2, imagePtMat.cols);
			Mat imgpti(imagePoints[i]);
			imgpti.copyTo(r);
		}
		fs << "Image_points" << imagePtMat;
	}
}

bool imageplus::kinect_calibration::runCalibrationAndSave(Settings& s,
		Size imageSize, Mat& cameraMatrix, Mat& distCoeffs,
		vector<vector<Point2f> > imagePoints) {
	vector<Mat> rvecs, tvecs;
	vector<float> reprojErrs;
	double totalAvgErr = 0;

	cout << "dins:runcalibration abans" << endl;
	bool ok = runCalibration(s, imageSize, cameraMatrix, distCoeffs,
			imagePoints, rvecs, tvecs, reprojErrs, totalAvgErr);

	cout << "dins:runcalibration després" << endl;
	cout << (ok ? "Calibration succeeded" : "Calibration failed")
			<< ". avg re projection error = " << totalAvgErr;

	if (ok)
		saveCameraParams(s, imageSize, cameraMatrix, distCoeffs, rvecs, tvecs,
				reprojErrs, imagePoints, totalAvgErr);
	return ok;
}

bool imageplus::kinect_calibration::runCalibrationExtrinsicsAndSave(Settings& s,
		Size imageSize, Mat cameraMatrix, Mat distCoeffs,
		vector<vector<Point2f> > imagePoints) {
	Mat rvecs, tvecs;
	vector<float> reprojErrs;
	double totalAvgErr = 0;

	bool ok = runCalibration_extrinsics(s, imageSize, cameraMatrix, distCoeffs,
			imagePoints, rvecs, tvecs, reprojErrs, totalAvgErr);
	cout
			<< (ok ?
					"Calibration extrinsics succeeded" :
					"Calibration extrinsics failed")
			<< ". avg re projection error = " << totalAvgErr;

	if (ok)
		saveCameraParams_extrinsics(s, imageSize, cameraMatrix, distCoeffs,
				rvecs, tvecs, reprojErrs, imagePoints, totalAvgErr);
	return ok;
}

pcl::PointCloud<pcl::PointXYZRGB>::Ptr imageplus::kinect_register_reconstruction::transform_pointcloud_extrinsics(
		const pcl::PointCloud<pcl::PointXYZRGB>::ConstPtr &source_cloud,
		Eigen::Matrix4d &extrinsics_matrix, bool asus) {

	/* Reminder: how transformation matrices work :

	 |-------> This column is the translation
	 | 1 0 0 x |  \
   | 0 1 0 y |   }-> The identity 3x3 matrix (no rotation) on the left
	 | 0 0 1 z |  /
	 | 0 0 0 1 |    -> We do not use this line (and it has to stay 0,0,0,1)

	 METHOD #1: Using a Matrix4f
	 This is the "manual" method, perfect to understand but error prone !
	 */
	Eigen::Matrix4f transform_1 = Eigen::Matrix4f::Identity();

	// Define a rotation matrix (see https://en.wikipedia.org/wiki/Rotation_matrix)
	float theta = M_PI / 4; // The angle of rotation in radians
	transform_1(0, 0) = -1;
	transform_1(1, 1) = -1;
	transform_1(2, 2) = 1;

	//    (row, column)

	// Define a translation of 2.5 meters on the x axis.
	transform_1(0, 3) = 5 * 0.107;
	transform_1(1, 3) = 7 * 0.107;

	// Print the transformation

	// Executing the transformation
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr transformed_cloud(
			new pcl::PointCloud<pcl::PointXYZRGB>());

	if (asus) {

		pcl::PointCloud<pcl::PointXYZRGB>::Ptr extrinsics_cloud(
				new pcl::PointCloud<pcl::PointXYZRGB>());
		// You can either apply transform_1 or transform_2; they are the same

		pcl::transformPointCloud(*source_cloud, *extrinsics_cloud,
				extrinsics_matrix);
		transformPointCloud(*extrinsics_cloud, *transformed_cloud, transform_1);
	} else {

		pcl::transformPointCloud(*source_cloud, *transformed_cloud,
				extrinsics_matrix);

	}

	return transformed_cloud;
}

void tools::build_pointCloud(Mat& depth,
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud, Mat& cameraMatrix,
		Mat& distCoeffs, int pc_color) {

	Size imageSize = depth.size();
	Mat depth_m, depth_und, map1, map2;

	initUndistortRectifyMap(cameraMatrix, distCoeffs, Mat(),
			getOptimalNewCameraMatrix(cameraMatrix, distCoeffs, imageSize, 1,
					imageSize, 0), imageSize, CV_16SC2, map1, map2);
	remap(depth, depth_und, map1, map2, INTER_NEAREST);
	//undistorted image. not working right now
	depth_und.convertTo(depth_m, CV_32F, 0.001);
	// depth.convertTo(depth_m, CV_32F, 0.001 );

	pcl::PointXYZRGB point_pcl;
	//   pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZRGB>);

	//test//
	int rgb_x, rgb_y;
	Vec3b color;

	float fx_ir = cameraMatrix.at<double>(0, 0);
	float fy_ir = cameraMatrix.at<double>(1, 1);

	float cx_ir = cameraMatrix.at<double>(0, 2);
	float cy_ir = cameraMatrix.at<double>(1, 2);

	for (int x = 0; x < depth_m.cols; x++) {

		for (int y = 0; y < depth_m.rows; y++) {

			if (depth_m.at<float>(y, x) != 0) {

				point_pcl.x = (float) (((x - cx_ir) * (depth_m.at<float>(y, x)))
						/ fx_ir);
				point_pcl.y = (float) (((y - cy_ir) * (depth_m.at<float>(y, x)))
						/ fy_ir);
				point_pcl.z = depth_m.at<float>(y, x);
					if (pc_color == 1) {
					point_pcl.r = 0;
					point_pcl.g = 255;
					point_pcl.b = 0;
				} else {
					point_pcl.r = 255;
					point_pcl.g = 0;
					point_pcl.b = 0;

				}
				cloud->points.push_back(point_pcl);
			}

		}
	}
}
void tools::build_pointCloud_RGB(Mat& depth, Mat& image,
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud, Mat& cameraMatrix,
		Mat& cameraMatrixRGB, Mat& distCoeffs, Mat& distCoeffsRGB,
		Mat& reg_rotation, Mat& reg_translation) {
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_rgb(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	Size imageSize = depth.size();
	Size imageSizeRGB = image.size();
	Mat depth_m, depth_und, image_und, map1, map2;

	initUndistortRectifyMap(cameraMatrix, distCoeffs, Mat(),
			getOptimalNewCameraMatrix(cameraMatrix, distCoeffs, imageSize, 1,
					imageSize, 0), imageSize, CV_16SC2, map1, map2);
	remap(depth, depth_und, map1, map2, INTER_NEAREST);

	depth_und.convertTo(depth_m, CV_32F, 0.001);

	pcl::PointXYZRGB point_pcl;

	int rgb_x, rgb_y;
	Vec3b color;

	float fx_ir = cameraMatrix.at<double>(0, 0);
	float fy_ir = cameraMatrix.at<double>(1, 1);

	float cx_ir = cameraMatrix.at<double>(0, 2);
	float cy_ir = cameraMatrix.at<double>(1, 2);

	vector < Point3f > ir_points;
	vector < Point2f > rgb_points;

	double vector[4] = { 0, 0, 0, 1 };
	Mat vector_cv(1, 4, CV_64F, vector);
	Mat t_reg_rotation, t_reg_rotation_def, reg_matrix_original;
	cv::transpose(reg_rotation, t_reg_rotation);
	cv::transpose(reg_rotation, t_reg_rotation_def);
	Mat t_reg_translation = -t_reg_rotation * reg_translation;

	hconcat(t_reg_rotation, t_reg_translation, t_reg_rotation);
	hconcat(reg_rotation, reg_translation, reg_matrix_original);
	vconcat(t_reg_rotation, vector_cv, t_reg_rotation);
	vconcat(reg_matrix_original, vector_cv, reg_matrix_original);
	Eigen::Map<
			Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic,
					Eigen::RowMajor>> registration_matrix_(
			t_reg_rotation.ptr<double>(), t_reg_rotation.rows,
			t_reg_rotation.cols);
	Eigen::Matrix4d registration_matrix = registration_matrix_;
	Eigen::Map<
			Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic,
					Eigen::RowMajor>> registration_matrix_original_(
			reg_matrix_original.ptr<double>(), reg_matrix_original.rows,
			reg_matrix_original.cols);
	Eigen::Matrix4d registration_matrix_original = registration_matrix_original_;

	for (int x = 0; x < depth_m.cols; x++) {

		for (int y = 0; y < depth_m.rows; y++) {

			if (depth_m.at<float>(y, x) != 0) {

				point_pcl.x = (float) (((x - cx_ir) * (depth_m.at<float>(y, x)))
						/ fx_ir);
				point_pcl.y = (float) (((y - cy_ir) * (depth_m.at<float>(y, x)))
						/ fy_ir);
				point_pcl.z = depth_m.at<float>(y, x);
			cloud->points.push_back(point_pcl);
			}

		}
	}
	pcl::transformPointCloud(*cloud, *cloud_rgb, registration_matrix);

	for (unsigned int i = 0; i < cloud_rgb->points.size(); i++) {
		ir_points.push_back(
				Point3f(cloud_rgb->points[i].x, cloud_rgb->points[i].y,
						cloud_rgb->points[i].z));

	}
	Mat rvec = Mat::zeros(3, 1, CV_64F);
	Mat tvec = Mat::zeros(3, 1, CV_64F);
	projectPoints(ir_points, rvec, tvec, cameraMatrixRGB, distCoeffsRGB,
			rgb_points);

	for (unsigned int i = 0; i < cloud->points.size(); i++) {
		if (rgb_points[i].x > 0 && rgb_points[i].y < 480 && rgb_points[i].y > 0
				&& rgb_points[i].x < 640) {
			Point3_ < uchar > *p = image.ptr < Point3_<uchar>
					> (rgb_points[i].y, rgb_points[i].x);
			cloud->points[i].b = p->x;
			cloud->points[i].g = p->y;
			cloud->points[i].r = p->z;
		} else {
			cloud->points[i].b = 0;
			cloud->points[i].g = 0;
			cloud->points[i].r = 255;
		}
	}
}

void kinect_register_reconstruction::scene_reconstruction_ownPC(
		std::string in_file, std::string in_file05, std::string inRGB01_file,
		std::string inRGB05_file, std::string ex_file, std::string ex_file05,
		std::string depth01_file, std::string depth05_file,
		std::string image01_file, std::string image05_file)

		{
	cout << " Option: Extrinsics" << endl;
	Mat distCoeffs, distCoeffsRGB01, distCoeffsRGB05, rvecs, tvecs,
			cameraMatrix, cameraMatrixRGB01, cameraMatrixRGB05, distCoeffs05,
			rvecs05, tvecs05, cameraMatrix05, reg_rot, reg_trans, reg_rot05,
			reg_trans05;

	//retrieving intrinsic and extrinsic parameters
	FileStorage in(in_file, FileStorage::READ);

	if (!in.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	in["Camera_Matrix"] >> cameraMatrix;
	in["Distortion_Coefficients"] >> distCoeffs;
	in.release();
	FileStorage inRGB01(inRGB01_file, FileStorage::READ);

	if (!inRGB01.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	inRGB01["Camera_Matrix"] >> cameraMatrixRGB01;
	inRGB01["Distortion_Coefficients"] >> distCoeffsRGB01;
	inRGB01["Reg_rotation"] >> reg_rot;
	inRGB01["Reg_trans"] >> reg_trans;
	inRGB01.release();

	FileStorage inRGB05(inRGB05_file, FileStorage::READ);

	if (!inRGB05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	inRGB05["Camera_Matrix"] >> cameraMatrixRGB05;
	inRGB05["Distortion_Coefficients"] >> distCoeffsRGB05;
	inRGB05["Reg_rotation"] >> reg_rot05;
	inRGB05["Reg_trans"] >> reg_trans05;
	inRGB05.release();

	FileStorage ex(ex_file, FileStorage::READ);
	if (!ex.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	ex["rvecs"] >> rvecs;
	ex["tvecs"] >> tvecs;
	ex.release();

	FileStorage in05(in_file05, FileStorage::READ);

	if (!in05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	in05["Camera_Matrix"] >> cameraMatrix05;
	in05["Distortion_Coefficients"] >> distCoeffs05;
	in05.release();

	FileStorage ex05(ex_file05, FileStorage::READ);
	if (!ex05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	ex05["rvecs"] >> rvecs05;
	ex05["tvecs"] >> tvecs05;
	ex05.release();

	Mat depth = imread(depth01_file, -1);
	Mat depth05 = imread(depth05_file, -1);
	Mat image = imread(image01_file);
	Mat image05 = imread(image05_file);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	//   tools::build_pointCloud(depth, cloud01, cameraMatrix, distCoeffs, 1);
	tools::build_pointCloud_RGB(depth, image, cloud01, cameraMatrix,
			cameraMatrixRGB01, distCoeffs, distCoeffsRGB01, reg_rot, reg_trans);

	// tools::build_pointCloud(depth05,cloud05,cameraMatrix05,distCoeffs05,2);
	tools::build_pointCloud_RGB(depth05, image05, cloud05, cameraMatrix05,
			cameraMatrixRGB05, distCoeffs05, distCoeffsRGB05, reg_rot05,
			reg_trans05);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05_cut(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_cut(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	double vector[4] = { 0, 0, 0, 1 };
	Mat vector_cv(1, 4, CV_64F, vector);

	Mat t_trans;
	Mat rotation_matrix, t_rotation_matrix;

	Rodrigues(rvecs, rotation_matrix);
	//get inverse of rotation matrix
	cv::transpose(rotation_matrix, t_rotation_matrix);
	// get inverse of translation vector
	t_trans = -t_rotation_matrix * tvecs;

	hconcat(t_rotation_matrix, t_trans, t_rotation_matrix);
	vconcat(t_rotation_matrix, vector_cv, t_rotation_matrix);

	Eigen::Map<
			Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic,
					Eigen::RowMajor>> extrinsics_matrix(
			t_rotation_matrix.ptr<double>(), t_rotation_matrix.rows,
			t_rotation_matrix.cols);
	Eigen::Matrix4d ex_m = extrinsics_matrix;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr reconstructed_cloud01 =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud01, ex_m, false);

	Mat t_trans05;
	Mat rotation_matrix05, t_rotation_matrix05;

	Rodrigues(rvecs05, rotation_matrix05);
	//get inverse of rotation matrix
	cv::transpose(rotation_matrix05, t_rotation_matrix05);
	// get inverse of translation vector
	t_trans05 = -t_rotation_matrix05 * tvecs05;

	hconcat(t_rotation_matrix05, t_trans05, t_rotation_matrix05);
	vconcat(t_rotation_matrix05, vector_cv, t_rotation_matrix05);

	Eigen::Map<
			Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic,
					Eigen::RowMajor>> extrinsics_matrix05(
			t_rotation_matrix05.ptr<double>(), t_rotation_matrix05.rows,
			t_rotation_matrix05.cols);
	Eigen::Matrix4d ex_m05 = extrinsics_matrix05;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr reconstructed_cloud05 =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud05, ex_m05, false);

	//ICP


	pcl::IterativeClosestPoint < pcl::PointXYZRGB, pcl::PointXYZRGB > icp;

	icp.setInputCloud(reconstructed_cloud01);
	icp.setInputTarget(reconstructed_cloud05);

	icp.setMaxCorrespondenceDistance(0.05);

	pcl::PointCloud < pcl::PointXYZRGB > icp_reconstructed_cloud01;

	icp.align(icp_reconstructed_cloud01);


	pcl::PointCloud<pcl::PointXYZRGB>::Ptr ptr_icp_reconstructed_cloud01(
			&icp_reconstructed_cloud01);


	std::cout << "has converged:" << icp.hasConverged() << " score: " <<

	icp.getFitnessScore() << std::endl;

	std::cout << icp.getFinalTransformation() << std::endl;

	std::vector<int> indices;
	pcl::removeNaNFromPointCloud(*reconstructed_cloud01, *reconstructed_cloud01,
			indices);
	pcl::removeNaNFromPointCloud(*reconstructed_cloud05, *reconstructed_cloud05,
			indices);
	pcl::removeNaNFromPointCloud(*ptr_icp_reconstructed_cloud01,
			*ptr_icp_reconstructed_cloud01, indices);


	for (unsigned int i = 0; i < ptr_icp_reconstructed_cloud01->points.size();
			i++) {
		ptr_icp_reconstructed_cloud01->points[i].r = 0;
		ptr_icp_reconstructed_cloud01->points[i].g = 255;
		ptr_icp_reconstructed_cloud01->points[i].b = 0;
	}

	for (unsigned int i = 0; i < reconstructed_cloud05->points.size(); i++) {
		reconstructed_cloud05->points[i].r = 255;
		reconstructed_cloud05->points[i].g = 0;
		reconstructed_cloud05->points[i].b = 0;
	}

	pcl::io::savePLYFile(
			"/imatge/acasanova/share/reconstruction_extrinsics+ICP_cloud01.ply",
			*reconstructed_cloud01);
	pcl::io::savePLYFile(
			"/imatge/acasanova/share/reconstruction_extrinsics+ICP_cloud05.ply",
			*reconstructed_cloud05);
	pcl::io::savePLYFile(
			"/imatge/acasanova/share/reconstruction_extrinsics+ICP_cloud01ICP.ply",
			*ptr_icp_reconstructed_cloud01);

	pcl::visualization::PCLVisualizer viewer("ICP reconstruction");

	viewer.addPointCloud(reconstructed_cloud01, "cloud");

	viewer.addCoordinateSystem(1.0, "cloud", 0);
	viewer.setBackgroundColor(0.05, 0.05, 0.05, 0); // Setting background to a dark grey
	//  viewer.setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 2, "cloud");

	while (!viewer.wasStopped()) { // Display the visualiser until 'q' key is pressed
		viewer.spinOnce();
	}
	cout << "end function" << endl;

}
/*
void tools::intersection_pyrmidalCut_extrinsics(
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud05,
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud01_RT, Mat& extr_05,
		Mat& extr_01, const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud05_cut,
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud01_RT_cut) {

	double optical_center[4] = { 0, 0, 0, 1 };
	Mat optical_center_cv(4, 1, CV_64F, optical_center);
	double optical_center_inf[4] = { 0, 0, 5, 1 };
	Mat optical_center_inf_cv(4, 1, CV_64F, optical_center_inf);
	Mat optical_center_05 = extr_05 * optical_center_cv;
	Mat optical_center_05_inf = extr_05 * optical_center_inf_cv;
	float angle = 57 / 2 * 3.14 / 180;
	//reference cloud
	float A, B, C, Ab, Bb, Cb;
	float r_line_x, r_line_y, r_line_xb, r_line_yb;

	r_line_x = optical_center_05.at<double>(0, 0)
			+ ((optical_center_05_inf.at<double>(0, 0)
					- optical_center_05.at<double>(0, 0)) * cos(angle))
			- ((optical_center_05_inf.at<double>(1, 0)
					- optical_center_05.at<double>(1, 0)) * sin(angle));
	r_line_y = optical_center_05.at<double>(1, 0)
			+ ((optical_center_05_inf.at<double>(0, 0)
					- optical_center_05.at<double>(0, 0)) * sin(angle))
			+ ((optical_center_05_inf.at<double>(1, 0)
					- optical_center_05.at<double>(1, 0)) * cos(angle));
	r_line_xb = optical_center_05.at<double>(0, 0)
			+ ((optical_center_05_inf.at<double>(0, 0)
					- optical_center_05.at<double>(0, 0)) * cos(angle))
			+ ((optical_center_05_inf.at<double>(1, 0)
					- optical_center_05.at<double>(1, 0)) * sin(angle));
	r_line_yb = optical_center_05.at<double>(1, 0)
			- ((optical_center_05_inf.at<double>(0, 0)
					- optical_center_05.at<double>(0, 0)) * sin(angle))
			+ ((optical_center_05_inf.at<double>(1, 0)
					- optical_center_05.at<double>(1, 0)) * cos(angle));
	A = optical_center_05.at<double>(1, 0) - r_line_y;
	B = r_line_x - optical_center_05.at<double>(0, 0);
	C = (optical_center_05.at<double>(0, 0) - r_line_x)
			* optical_center_05.at<double>(1, 0)
			+ (r_line_y - optical_center_05.at<double>(1, 0))
					* optical_center_05.at<double>(0, 0);
	Ab = optical_center_05.at<double>(1, 0) - r_line_yb;
	Bb = r_line_xb - optical_center_05.at<double>(0, 0);
	Cb = (optical_center_05.at<double>(0, 0) - r_line_xb)
			* optical_center_05.at<double>(1, 0)
			+ (r_line_yb - optical_center_05.at<double>(1, 0))
					* optical_center_05.at<double>(0, 0);

	float theta = atan(B / A);
	float p = -C / (sqrt(A * A + B * B));
	float theta_b = atan(Bb / Ab);
	float p_b = -Cb / (sqrt(Ab * Ab + Bb * Bb));

	//RT cloud
	Mat optical_center_01 = extr_01 * optical_center_cv;
	Mat optical_center_01_inf = extr_01 * optical_center_inf_cv;

	float A1, B1, C1, A1b, B1b, C1b;
	float r_line_x1, r_line_y1, r_line_x1b, r_line_y1b;
	r_line_x1 = optical_center_01.at<double>(0, 0)
			+ ((optical_center_01_inf.at<double>(0, 0)
					- optical_center_01.at<double>(0, 0)) * cos(angle))
			- ((optical_center_01_inf.at<double>(1, 0)
					- optical_center_01.at<double>(1, 0)) * sin(angle));
	r_line_y1 = optical_center_01.at<double>(1, 0)
			+ ((optical_center_01_inf.at<double>(0, 0)
					- optical_center_01.at<double>(0, 0)) * sin(angle))
			+ ((optical_center_01_inf.at<double>(1, 0)
					- optical_center_01.at<double>(1, 0)) * cos(angle));
	r_line_x1b = optical_center_01.at<double>(0, 0)
			+ ((optical_center_01_inf.at<double>(0, 0)
					- optical_center_01.at<double>(0, 0)) * cos(angle))
			+ ((optical_center_01_inf.at<double>(1, 0)
					- optical_center_01.at<double>(1, 0)) * sin(angle));
	r_line_y1b = optical_center_01.at<double>(1, 0)
			- ((optical_center_01_inf.at<double>(0, 0)
					- optical_center_01.at<double>(0, 0)) * sin(angle))
			+ ((optical_center_01_inf.at<double>(1, 0)
					- optical_center_01.at<double>(1, 0)) * cos(angle));
	A1 = optical_center_01.at<double>(1, 0) - r_line_y1;
	B1 = r_line_x1 - optical_center_01.at<double>(0, 0);
	C1 = (optical_center_01.at<double>(0, 0) - r_line_x1)
			* optical_center_01.at<double>(1, 0)
			+ (r_line_y1 - optical_center_01.at<double>(1, 0))
					* optical_center_01.at<double>(0, 0);
	A1b = optical_center_01.at<double>(1, 0) - r_line_y1b;
	B1b = r_line_x1b - optical_center_01.at<double>(0, 0);
	C1b = (optical_center_01.at<double>(0, 0) - r_line_x1b)
			* optical_center_01.at<double>(1, 0)
			+ (r_line_y1b - optical_center_01.at<double>(1, 0))
					* optical_center_01.at<double>(0, 0);

	float theta1 = atan(B1 / A1);
	float p1 = -C1 / (sqrt(A1 * A1 + B1 * B1));
	float theta_1b = atan(B1b / A1b);
	float p_1b = -C1b / (sqrt(A1b * A1b + B1b * B1b));

	for (unsigned int i = 0; i < cloud05->points.size(); i++) {

		if (cloud05->points[i].x * cos(theta_b)
				+ cloud05->points[i].y * sin(theta_b) + p_b < 0
				&& cloud05->points[i].x * cos(theta)
						+ cloud05->points[i].y * sin(theta) - p < 0) {
			if (cloud05->points[i].x * cos(theta_1b)
					+ cloud05->points[i].y * sin(theta_1b) - p_1b > 0
					&& cloud05->points[i].x * cos(theta1)
							+ cloud05->points[i].y * sin(theta1) - p1 < 0) {
				if ((cloud05->points[i].x < 1 && cloud05->points[i].x > 0)
						|| cloud05->points[i].x < -1)
					cloud05_cut->points.push_back(cloud05->points[i]);
			}
		}
	}

	for (unsigned int i = 0; i < cloud01_RT->points.size(); i++) {

		if (cloud01_RT->points[i].x * cos(theta_1b)
				+ cloud01_RT->points[i].y * sin(theta_1b) - p_1b > 0
				&& cloud01_RT->points[i].x * cos(theta1)
						+ cloud01_RT->points[i].y * sin(theta1) - p1 < 0) {
			if (cloud01_RT->points[i].x * cos(theta_b)
					+ cloud01_RT->points[i].y * sin(theta_b) + p_b < 0
					&& cloud01_RT->points[i].x * cos(theta)
							+ cloud01_RT->points[i].y * sin(theta) - p < 0) {
				if ((cloud01_RT->points[i].x < 1 && cloud01_RT->points[i].x > 0)
						|| cloud01_RT->points[i].x < -1)
					cloud01_RT_cut->points.push_back(cloud01_RT->points[i]);
			}
		}
	}

}
*/
void application::fingers_RT(std::string in01_file, std::string in05_file,
		std::string ex05_file, std::string inRGB01_file,
		std::string inRGB05_file, vector<std::string> img_list01,
		vector<std::string> img_list05, vector<Point2f> annoted_fingers01,
		vector<Point2f> annoted_fingers05, std::string depth_file,
		std::string depth05_file, std::string image01_file,
		std::string image05_file, int method) {

	if (method == 0)
		cout << " Option: figers_RT SVD" << endl;
	else
		cout << " Option: figers_RT ICP" << endl;
	Mat distCoeffs, distCoeffsRGB01, distCoeffsRGB05, rvecs, tvecs,
			cameraMatrix, cameraMatrixRGB01, cameraMatrixRGB05, distCoeffs05,
			rvecs05, tvecs05, cameraMatrix05, reg_rot, reg_trans, reg_rot05,
			reg_trans05;

	//retreiving intrinsic and extrinsic parameters
	FileStorage in(in01_file, FileStorage::READ);

	if (!in.isOpened())
		cout << "Could not open the configuration file: adsfasdf? \\" << endl;

	in["Camera_Matrix"] >> cameraMatrix;
	in["Distortion_Coefficients"] >> distCoeffs;
	in.release();

	FileStorage in05(in05_file, FileStorage::READ);

	if (!in05.isOpened())
		cout << "Could not open the configuration file:asdasd! \\" << endl;

	in05["Camera_Matrix"] >> cameraMatrix05;
	in05["Distortion_Coefficients"] >> distCoeffs05;
	in05.release();

	FileStorage inRGB01(inRGB01_file, FileStorage::READ);

	if (!inRGB01.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	inRGB01["Camera_Matrix"] >> cameraMatrixRGB01;
	inRGB01["Distortion_Coefficients"] >> distCoeffsRGB01;
	inRGB01["Reg_rotation"] >> reg_rot;
	inRGB01["Reg_trans"] >> reg_trans;
	inRGB01.release();

	FileStorage inRGB05(inRGB05_file, FileStorage::READ);

	if (!inRGB05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	inRGB05["Camera_Matrix"] >> cameraMatrixRGB05;
	inRGB05["Distortion_Coefficients"] >> distCoeffsRGB05;
	inRGB05["Reg_rotation"] >> reg_rot05;
	inRGB05["Reg_trans"] >> reg_trans05;
	inRGB05.release();

	FileStorage ex05(ex05_file, FileStorage::READ);
	if (!ex05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	ex05["rvecs"] >> rvecs05;
	ex05["tvecs"] >> tvecs05;
	ex05.release();

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	float fx_ir = cameraMatrix.at<double>(0, 0);
	float fy_ir = cameraMatrix.at<double>(1, 1);

	float cx_ir = cameraMatrix.at<double>(0, 2);
	float cy_ir = cameraMatrix.at<double>(1, 2);

	pcl::PointXYZRGB point_pcl;

	int pixel_x, pixel_y;
	Mat depth, depth_m;
	for (unsigned int i = 0; i < img_list01.size(); i++) {

		depth = imread(img_list01[i], -1);
		depth_m;
		depth.convertTo(depth_m, CV_32F, 0.001);
		pixel_x = annoted_fingers01[i].x;
		pixel_y = annoted_fingers01[i].y;
		point_pcl.x = (float) (((pixel_x - cx_ir)
				* (depth_m.at<float>(pixel_y, pixel_x))) / fx_ir);
		point_pcl.y = (float) (((pixel_y - cy_ir)
				* (depth_m.at<float>(pixel_y, pixel_x))) / fy_ir);
		point_pcl.z = depth_m.at<float>(pixel_y, pixel_x);
		point_pcl.r = 0;
		point_pcl.g = 255;
		point_pcl.b = 0;
		cloud01->points.push_back(point_pcl);

	}
	for (unsigned int i = 0; i < img_list05.size(); i++) {

		depth = imread(img_list05[i], -1);
		depth_m;
		depth.convertTo(depth_m, CV_32F, 0.001);
		pixel_x = annoted_fingers05[i].x;
		pixel_y = annoted_fingers05[i].y;
		point_pcl.x = (float) (((pixel_x - cx_ir)
				* (depth_m.at<float>(pixel_y, pixel_x))) / fx_ir);
		point_pcl.y = (float) (((pixel_y - cy_ir)
				* (depth_m.at<float>(pixel_y, pixel_x))) / fy_ir);
		point_pcl.z = depth_m.at<float>(pixel_y, pixel_x);
		point_pcl.r = 255;
		point_pcl.g = 0;
		point_pcl.b = 0;
		cloud05->points.push_back(point_pcl);

	}
//passing points to vector

	vector<Point3f> v_cloud01, v_cloud05, v_cloud01_RT;
	for (unsigned int i = 0; i < cloud01->points.size(); i++) {
		v_cloud01.push_back(
				Point3f(cloud01->points[i].x, cloud01->points[i].y,
						cloud01->points[i].z));
	}
	for (unsigned int i = 0; i < cloud05->points.size(); i++) {
		v_cloud05.push_back(
				Point3f(cloud05->points[i].x, cloud05->points[i].y,
						cloud05->points[i].z));
	}

//estimating RT
	//   Mat estRT,inliers;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_RT(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	pcl::registration::TransformationEstimationSVD<pcl::PointXYZRGB,
			pcl::PointXYZRGB>::Matrix4 estRT;
	if (method == 0) {
		pcl::registration::TransformationEstimationSVD < pcl::PointXYZRGB, pcl::PointXYZRGB
				> TESVD;
		TESVD.estimateRigidTransformation(*cloud01, *cloud05, estRT);
		cout << "SVD RT matrix " << estRT << endl;

		transformPointCloud(*cloud01, *cloud01_RT, estRT);

		for (unsigned int i = 0; i < cloud01_RT->points.size(); i++) {
			v_cloud01_RT.push_back(
					Point3f(cloud01_RT->points[i].x, cloud01_RT->points[i].y,
							cloud01_RT->points[i].z));
		}

		//error distance
		double dist = (norm(v_cloud05, v_cloud01_RT, NORM_L1))
				/ v_cloud05.size();

		cout << "MSE error: " << dist << endl;
	}
	Eigen::Matrix4d icp_RT_def;
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_RT_icp_def;
	if (method == 1) {
		pcl::IterativeClosestPoint < pcl::PointXYZRGB, pcl::PointXYZRGB > icp;

		icp.setInputCloud(cloud01);
		icp.setInputTarget(cloud05);

		//   icp.setMaxCorrespondenceDistance (0.1);

		icp.setTransformationEpsilon(1e-7);
		icp.setEuclideanFitnessEpsilon(1e-5);

		pcl::PointCloud < pcl::PointXYZRGB > cloud01_icp;
		icp.align(cloud01_icp);

		Eigen::Matrix4f icp_RT = icp.getFinalTransformation();
		icp_RT_def = icp_RT.cast<double>();
		cout << icp_RT_def << endl;
		pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_RT_icp(&cloud01_icp);

		cloud01_RT_icp_def = cloud01_RT_icp;
		transformPointCloud(*cloud01, *cloud01_RT, icp_RT_def);


	}


	depth = imread(depth_file, -1);
	Mat depth05 = imread(depth05_file, -1);
	Mat image = imread(image01_file);
	Mat image05 = imread(image05_file);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05_scene(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_RT(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	// tools::build_pointCloud(depth, cloud01_scene, cameraMatrix, distCoeffs, 1);
	tools::build_pointCloud_RGB(depth, image, cloud01_scene, cameraMatrix,
			cameraMatrixRGB01, distCoeffs, distCoeffsRGB01, reg_rot, reg_trans);

	//tools::build_pointCloud(depth05, cloud05_scene, cameraMatrix05, distCoeffs05, 2);
	tools::build_pointCloud_RGB(depth05, image05, cloud05_scene, cameraMatrix05,
			cameraMatrixRGB05, distCoeffs05, distCoeffsRGB05, reg_rot05,
			reg_trans05);

	if (method == 0)
		transformPointCloud(*cloud01_scene, *cloud01_scene_RT, estRT);
	else
		transformPointCloud(*cloud01_scene, *cloud01_scene_RT, icp_RT_def);


	pcl::IterativeClosestPoint < pcl::PointXYZRGB, pcl::PointXYZRGB > icp;

	icp.setInputCloud(cloud01_scene_RT);
	icp.setInputTarget(cloud05_scene);

	// Set the max correspondence distance to 1cm (e.g., correspondences with higher distances will be ignored)
	icp.setMaxCorrespondenceDistance(0.1);

	pcl::PointCloud < pcl::PointXYZRGB > cloud01_scene_RT_icp;

	icp.align(cloud01_scene_RT_icp);

	Eigen::Matrix4f icp_RT = icp.getFinalTransformation();

	Eigen::Matrix4d icp_RTdef = icp_RT.cast<double>();
	cout << icp_RTdef << endl;


	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_RT_icp_ptr(
			&cloud01_scene_RT_icp);


	std::cout << "has converged:" << icp.hasConverged() << " score: " <<

	icp.getFitnessScore() << std::endl;

	std::cout << icp.getFinalTransformation() << std::endl;

	/*     for (unsigned int i = 0; i < cloud01_scene_RT_icp_ptr->points.size();
	 i++)
	 {
	 cloud01_scene_RT_icp_ptr->points[i].r = 0;
	 cloud01_scene_RT_icp_ptr->points[i].g = 255;
	 cloud01_scene_RT_icp_ptr->points[i].b = 255;
	 }*/

	//  *cloud01_scene+=*cloud05_scene;
	//   *cloud05_scene+=*cloud01_scene_RT;
	//   *cloud05_scene+=*cloud01_scene_RT_icp_ptr;
	*cloud01 += *cloud05;
	*cloud01 += *cloud01_RT;

	//*cloud01_scene+=*cloud01_scene_RT_icp_ptr;
	std::vector<int> indices;

	double vector[4] = { 0, 0, 0, 1 };
	Mat vector_cv(1, 4, CV_64F, vector);
	Mat t_trans05;
	Mat rotation_matrix05, t_rotation_matrix05;

	Rodrigues(rvecs05, rotation_matrix05);
	//get inverse of rotation matrix
	cv::transpose(rotation_matrix05, t_rotation_matrix05);
	// get inverse of translation vector
	t_trans05 = -t_rotation_matrix05 * tvecs05;

	hconcat(t_rotation_matrix05, t_trans05, t_rotation_matrix05);
	vconcat(t_rotation_matrix05, vector_cv, t_rotation_matrix05);

	Eigen::Map<
			Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic,
					Eigen::RowMajor>> extrinsics_matrix05(
			t_rotation_matrix05.ptr<double>(), t_rotation_matrix05.rows,
			t_rotation_matrix05.cols);
	Eigen::Matrix4d ex_m05 = extrinsics_matrix05;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05_scene_ex =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud05_scene, ex_m05, false);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_RT_ex =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud01_scene_RT, ex_m05, false);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_RT_icp_ptr_ex =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud01_scene_RT_icp_ptr, ex_m05, false);

	pcl::removeNaNFromPointCloud(*cloud05_scene, *cloud05_scene, indices);
	pcl::io::savePLYFile("/imatge/acasanova/share/cloud05_scene.ply",
			*cloud05_scene_ex);
	pcl::io::savePLYFile("/imatge/acasanova/share/cloud01_scene_RT.ply",
			*cloud01_scene_RT_ex);
	pcl::io::savePLYFile("/imatge/acasanova/share/cloud01_scene_RT_ICP.ply",
			*cloud01_scene_RT_icp_ptr_ex);
	pcl::io::savePLYFile("/imatge/acasanova/share/fingers_manual_detection.ply",
			*cloud01);

	pcl::visualization::PCLVisualizer viewer("fingers test");

	pcl::visualization::PointCloudColorHandlerCustom < pcl::PointXYZRGB
			> c(cloud01_scene_RT, 0, 0, 255);

	viewer.addPointCloud(cloud01, "cloud_f"); //green
	viewer.addCoordinateSystem(1.0, "cloud", 0);
	viewer.setBackgroundColor(0.05, 0.05, 0.05, 0); // Setting background to a dark grey

	while (!viewer.wasStopped()) { // Display the visualiser until 'q' key is pressed
		viewer.spinOnce();
	}

}

void application::fingers_RT_detector(std::string in_file,
		std::string in05_file, std::string ex05_file, std::string inRGB01_file,
		std::string inRGB05_file, std::string detected_fingers01,
		std::string detected_fingers05, std::string depth_file,
		std::string depth05_file, std::string image01_file,
		std::string image05_file) {

	cout << " Option: figers_RT_detector" << endl;

	Mat distCoeffs, distCoeffsRGB01, distCoeffsRGB05, rvecs, tvecs,
			cameraMatrix, cameraMatrixRGB01, cameraMatrixRGB05, distCoeffs05,
			rvecs05, tvecs05, cameraMatrix05, reg_rot, reg_trans, reg_rot05,
			reg_trans05;
	vector<Point3f> detected_coords, detected_coords05;

	//retreiving intrinsic and extrinsic parameters
	FileStorage in(in_file, FileStorage::READ);

	if (!in.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	in["Camera_Matrix"] >> cameraMatrix;
	in["Distortion_Coefficients"] >> distCoeffs;
	in.release();

	FileStorage in05(in05_file, FileStorage::READ);

	if (!in05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	in05["Camera_Matrix"] >> cameraMatrix05;
	in05["Distortion_Coefficients"] >> distCoeffs05;
	in05.release();

	FileStorage inRGB01(inRGB01_file, FileStorage::READ);

	if (!inRGB01.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	inRGB01["Camera_Matrix"] >> cameraMatrixRGB01;
	inRGB01["Distortion_Coefficients"] >> distCoeffsRGB01;
	inRGB01["Reg_rotation"] >> reg_rot;
	inRGB01["Reg_trans"] >> reg_trans;
	inRGB01.release();

	FileStorage inRGB05(inRGB05_file, FileStorage::READ);

	if (!inRGB05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	inRGB05["Camera_Matrix"] >> cameraMatrixRGB05;
	inRGB05["Distortion_Coefficients"] >> distCoeffsRGB05;
	inRGB05["Reg_rotation"] >> reg_rot05;
	inRGB05["Reg_trans"] >> reg_trans05;
	inRGB05.release();

	FileStorage ex05(ex05_file, FileStorage::READ);
	if (!ex05.isOpened())
		cout << "Could not open the configuration file: \\" << endl;

	ex05["rvecs"] >> rvecs05;
	ex05["tvecs"] >> tvecs05;
	ex05.release();

	FileStorage fingers01(detected_fingers01, FileStorage::READ);

	if (!fingers01.isOpened())
		cout << "Could not open fingers file: \\" << endl;

	fingers01["Detected_coords"] >> detected_coords;
	std::cout << "Fingers 01 " << detected_coords.size() << endl;
	fingers01.release();

	FileStorage fingers05(detected_fingers05, FileStorage::READ);

	if (!fingers05.isOpened())
		cout << "Could not open fingers file: \\" << endl;

	fingers05["Detected_coords"] >> detected_coords05;
	std::cout << "Fingers 05 " << detected_coords05.size() << endl;
	fingers05.release();
	cout << "Detected coords size " << detected_coords.size() << " "
			<< detected_coords05.size() << endl;
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_f(
			new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05_f(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	float fx_ir = cameraMatrix.at<double>(0, 0);
	float fy_ir = cameraMatrix.at<double>(1, 1);

	float cx_ir = cameraMatrix.at<double>(0, 2);
	float cy_ir = cameraMatrix.at<double>(1, 2);

	pcl::PointXYZRGB point_pcl;
	float pixel_x, pixel_y;
	float depth_m;
	unsigned int size;

	if (detected_coords.size() <= detected_coords05.size()) {
		size = detected_coords.size();
	} else {
		size = detected_coords05.size();
	}
	for (unsigned int i = 0; i < size; i++) {
		pixel_x = detected_coords[i].x;
		pixel_y = detected_coords[i].y;
		depth_m = detected_coords[i].z / 1000;

		point_pcl.x = (float) (((pixel_x - cx_ir) * (depth_m)) / fx_ir);
		point_pcl.y = (float) (((pixel_y - cy_ir) * (depth_m)) / fy_ir);
		point_pcl.z = depth_m;
		point_pcl.r = 0;
		point_pcl.g = 255;
		point_pcl.b = 0;

		cloud01->points.push_back(point_pcl);

		pixel_x = detected_coords05[i].x;
		pixel_y = detected_coords05[i].y;
		depth_m = detected_coords05[i].z / 1000;
		point_pcl.r = 255;
		point_pcl.g = 0;
		point_pcl.b = 0;

		point_pcl.x = (float) (((pixel_x - cx_ir) * (depth_m)) / fx_ir);
		point_pcl.y = (float) (((pixel_y - cy_ir) * (depth_m)) / fy_ir);



		point_pcl.z = depth_m;

		cloud05->points.push_back(point_pcl);

	}
	pcl::StatisticalOutlierRemoval < pcl::PointXYZRGB > sor;
	sor.setInputCloud(cloud01);
	sor.setMeanK(5);
	sor.setStddevMulThresh(1);
	sor.filter(*cloud01);
	sor.setInputCloud(cloud05);
	sor.filter(*cloud05);

	int size_f;

	if (cloud05->points.size() <= cloud01->points.size()) {
		size_f = cloud05->points.size();
	} else {
		size_f = cloud01->points.size();
	}
	for (int i = 0; i < size_f; i++) {

		cloud05_f->points.push_back(cloud05->points[i]);

		cloud01_f->points.push_back(cloud01->points[i]);
	}

	pcl::IterativeClosestPoint < pcl::PointXYZRGB, pcl::PointXYZRGB > icp;

	icp.setInputCloud(cloud01_f);
	icp.setInputTarget(cloud05_f);

	icp.setTransformationEpsilon(1e-7);
	icp.setEuclideanFitnessEpsilon(1e-5);

	pcl::PointCloud < pcl::PointXYZRGB > cloud01_icp;
	icp.align(cloud01_icp);

	Eigen::Matrix4f icp_RT = icp.getFinalTransformation();
	cout << "first tf" << icp_RT << endl;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_icp_ptr_(&cloud01_icp);

	icp.setInputCloud(cloud01_icp_ptr_);
	icp.setInputTarget(cloud05_f);

	icp.setMaxCorrespondenceDistance(0.05);
	icp.setMaximumIterations(2000);
	icp.setTransformationEpsilon(1e-10);
	icp.setEuclideanFitnessEpsilon(1e-10);

	pcl::PointCloud < pcl::PointXYZRGB > cloud01_icp_icp;
	icp.align(cloud01_icp_icp);

	Eigen::Matrix4f icp_RT2 = icp.getFinalTransformation();
	Eigen::Matrix4d icp_RTdef_ = icp_RT.cast<double>();
	Eigen::Matrix4d icp_RTdef_2 = icp_RT2.cast<double>();
	Eigen::Matrix4d icp_RTdef = icp_RTdef_2 * icp_RTdef_;
	cout << "second tf" << icp_RT2 << endl;
	cout << "TF fingers" << icp_RTdef << endl;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_icp_ptr(&cloud01_icp_icp);


	Mat depth = imread(depth_file, -1);
	Mat depth05 = imread(depth05_file, -1);

	Mat image = imread(image01_file);
	Mat image05 = imread(image05_file);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05_scene(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	// tools::build_pointCloud(depth, cloud01_scene, cameraMatrix, distCoeffs, 1);
	tools::build_pointCloud_RGB(depth, image, cloud01_scene, cameraMatrix,
			cameraMatrixRGB01, distCoeffs, distCoeffsRGB01, reg_rot, reg_trans);

	//  tools::build_pointCloud(depth05,cloud05_scene,cameraMatrix05,distCoeffs05,2);
	tools::build_pointCloud_RGB(depth05, image05, cloud05_scene, cameraMatrix05,
			cameraMatrixRGB05, distCoeffs05, distCoeffsRGB05, reg_rot05,
			reg_trans05);

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_RT(
			new pcl::PointCloud<pcl::PointXYZRGB>);

	transformPointCloud(*cloud01_scene, *cloud01_scene_RT, icp_RTdef);


	pcl::visualization::PCLVisualizer viewer("test");

	pcl::visualization::PointCloudColorHandlerCustom < pcl::PointXYZRGB
			> c05(cloud05, 255, 0, 0);
	pcl::visualization::PointCloudColorHandlerCustom < pcl::PointXYZRGB
			> c01(cloud01, 0, 255, 0);
	pcl::visualization::PointCloudColorHandlerCustom < pcl::PointXYZRGB
			> c01_RT(cloud01_icp_ptr, 0, 0, 255);

	viewer.addPointCloud(cloud01, c01, "cloud"); //red
	viewer.addPointCloud(cloud05, c05, "cloud05");
	viewer.addPointCloud(cloud01_icp_ptr, c01_RT, "cloud01_RT");
	viewer.addPointCloud(cloud01_scene_RT, "cloud01_scene_RT");
	viewer.addPointCloud(cloud05_scene, "cloud05_scene");

	viewer.addCoordinateSystem(1.0, "cloud", 0);
	viewer.setBackgroundColor(0.05, 0.05, 0.05, 0); // Setting background to a dark grey

	while (!viewer.wasStopped()) { // Display the visualiser until 'q' key is pressed
		viewer.spinOnce();
	}

	cout << "Applying ICP to transformed cloud" << endl;
	icp.setInputCloud(cloud01_scene_RT);
	icp.setInputTarget(cloud05_scene);

	// Set the max correspondence distance to 5cm (e.g., correspondences with higher distances will be ignored)
	icp.setMaxCorrespondenceDistance(0.05);

	icp.setTransformationEpsilon(1e-7);
	// Set the euclidean distance difference epsilon (criterion 3)
	icp.setEuclideanFitnessEpsilon(1e-5);

	pcl::PointCloud < pcl::PointXYZRGB > cloud01_scene_icp;
	icp.align(cloud01_scene_icp);

	Eigen::Matrix4f icp_RT_scene = icp.getFinalTransformation();
	Eigen::Matrix4d icp_RT_scene_def = icp_RT_scene.cast<double>();
	cout << "scene TF" << icp_RT_scene << endl;
	Eigen::Matrix4d finalTF = icp_RT_scene_def * icp_RTdef;
	cv::Mat finalTF_cv;
	cv::eigen2cv(finalTF, finalTF_cv);
	cout << "Final Transformation : " << finalTF << endl;
	cout << "Final Transformation : " << finalTF_cv << endl;

	cv::FileStorage fs("/imatge/acasanova/work/Fingers_finalTF.xml",
			cv::FileStorage::WRITE);

	if (!fs.isOpened())
		cout << "Could not open the configuration file: \\" << endl;
	fs << "FinalTF" << finalTF_cv;
	fs.release();

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_icp_ptr_(
			&cloud01_scene_icp);


	*cloud01 += *cloud01_icp_ptr;
	*cloud01 += *cloud05;

	// applying extr to mesure error
	double vector[4] = { 0, 0, 0, 1 };
	Mat vector_cv(1, 4, CV_64F, vector);
	Mat t_trans05;
	Mat rotation_matrix05, t_rotation_matrix05;

	Rodrigues(rvecs05, rotation_matrix05);
	//get inverse of rotation matrix
	cv::transpose(rotation_matrix05, t_rotation_matrix05);
	// get inverse of translation vector
	t_trans05 = -t_rotation_matrix05 * tvecs05;

	hconcat(t_rotation_matrix05, t_trans05, t_rotation_matrix05);
	vconcat(t_rotation_matrix05, vector_cv, t_rotation_matrix05);

	Eigen::Map<
			Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic,
					Eigen::RowMajor>> extrinsics_matrix05(
			t_rotation_matrix05.ptr<double>(), t_rotation_matrix05.rows,
			t_rotation_matrix05.cols);
	Eigen::Matrix4d ex_m05 = extrinsics_matrix05;

	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud05_scene_ex =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud05_scene, ex_m05, false);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_RT_ex =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud01_scene_RT, ex_m05, false);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01_scene_icp_ptr_ex =
			kinect_register_reconstruction::transform_pointcloud_extrinsics(
					cloud01_scene_icp_ptr_, ex_m05, false);

	pcl::io::savePLYFile(
			"/imatge/acasanova/share/detector_reco_scene_ICP_05_person.ply",
			*cloud05_scene);

	pcl::io::savePLYFile(
			"/imatge/acasanova/share/detector_reco_scene_ICP_01ICP_person.ply",
			*cloud01_scene_icp_ptr_);
	pcl::io::savePLYFile(
			"/imatge/acasanova/share/detector_reco_scene_ICP_01_person.ply",
			*cloud01_scene_RT);

	pcl::io::savePLYFile("/imatge/acasanova/share/detector_finger_clouds.ply",
			*cloud01);

	/*
	 pcl::visualization::PCLVisualizer viewer("test");

	 // pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZRGB> c05 (cloud05, 255, 0, 0);
	 // pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZRGB> c01 (cloud01, 0, 0, 255);
	 // pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZRGB> c01_RT (cloud01_RT, 0, 255, 0);
	 pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZRGB> c01_icp (cloud01_icp_ptr, 255, 255, 255);

	 viewer.addPointCloud(cloud01, "cloud"); //red
	 viewer.addPointCloud(cloud05, "cloud05");
	 //  viewer.addPointCloud(cloud01_RT, c01_RT, "cloud01_RT");

	 viewer.addPointCloud(cloud01_icp_ptr, c01_icp,"cloud01_icp");

	 viewer.addCoordinateSystem(1.0, "cloud", 0);
	 viewer.setBackgroundColor(0.05, 0.05, 0.05, 0); // Setting background to a dark grey
	 //      viewer.setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 7, "test");
	 //  viewer.setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 7, "center05");

	 while (!viewer.wasStopped())
	 { // Display the visualiser until 'q' key is pressed
	 viewer.spinOnce();
	 }
	 */
}

