/*
 * calibrate_RGB.hpp
 *
 *
 * Created on: Feb 15,2016
 * 		Author: Arantxa Casanova
 */


#ifndef IMAGEPLUS_CALIBRATE_RGB_HPP
#define IMAGEPLUS_CALIBRATE_RGB_HPP


#include <iostream>
#include <sstream>
#include <time.h>
#include <stdio.h>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <imageplus/kinect_calibration/calibrate_RGB.hpp>

#include <pcl/visualization/cloud_viewer.h>
#include <pcl/io/oni_grabber.h>
#include <boost/thread/thread.hpp>
#include <pcl/common/common_headers.h>
#include <pcl/common/transforms.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/io/pcd_io.h>
#include <pcl/io/ply_io.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/console/parse.h>
#include <pcl/filters/filter.h>
#include <pcl/point_types.h>
#include <opencv2/core/eigen.hpp>

#include <pcl/registration/icp.h>
#include <pcl/features/normal_3d.h>
#include <pcl/registration/icp_nl.h>
#include <pcl/registration/transforms.h>
#include <pcl/keypoints/sift_keypoint.h>
#include <pcl/features/fpfh.h>
#include <pcl/impl/point_types.hpp>
#include <pcl/registration/correspondence_estimation.h>
#include <pcl/registration/correspondence_rejection_sample_consensus.h>
#include <pcl/correspondence.h>
#include <pcl/filters/statistical_outlier_removal.h>
using namespace cv;
using namespace std;
using namespace pcl;

/*!@file
*/
//!  Settings class
/*!
  To read and write .XML files containing calibration parameters.
*/
class Settings{

       public:
           Settings() : goodInput(false) {}
           enum Pattern { NOT_EXISTING, CHESSBOARD, CIRCLES_GRID, ASYMMETRIC_CIRCLES_GRID }; /*!< Enum value Pattern. */
           enum InputType {INVALID, CAMERA, VIDEO_FILE, IMAGE_LIST};/*!< Enum value InputType. */
           void read(const FileNode& node);
           void write(FileStorage& file) const;
           void interprate();
           Mat nextImage(int& atImageList, vector<string> imageList);
           bool readStringList( const string& filename, vector<string>& l );


       public:
           Size boardSize_in;            //! The size of the intrinsics board -> Number of items by width and height
           Size boardSize_ex;            //! The size of the extrinsics board -> Number of items by width and height

           Pattern calibrationPattern; //! One of the Chessboard, circles, or asymmetric circle pattern
           float squareSize_in;          //! The size of a square (intrinsics pattern) in your defined unit (point, millimeter,etc).
           float squareSize_ex;          //! The size of a square (extrinsics pattern) in your defined unit (point, millimeter,etc).
           int nrFrames;              //! The number of frames to use from the input for calibration
           float aspectRatio;         //! The aspect ratio
           int delay;                 //! In case of a video input
           bool bwritePoints;        //!  Write detected feature points
           bool bwriteExtrinsics;     //! Write extrinsic parameters
           bool calibZeroTangentDist; //! Assume zero tangential distortion
           bool calibFixPrincipalPoint;//! Fix the principal point at the center
           bool flipVertical;          //! Flip the captured images around the horizontal axis
           string outputFileName_in;      //! The name of the file where to write intrinsics parameters
           string outputFileName_ex;      //! The name of the file where to write extrinsics parameters
           string cornersFileName_ex;	//! File where position of corners (in pixels) is stored
           bool showUndistorsed;       //! Show undistorted images after calibration
           string input_in;              //! Input  images to calculate intrinsics
           string input_ex;              //! Input images to calculate extrinsics



           int cameraID;
           vector<string> imageList_in;
           vector<string> imageList_ex;
           int atImageList_in;
           int atImageList_ex;
           VideoCapture inputCapture;
           InputType inputType;
           bool goodInput;
           int flag;

       private:
           string patternToUse;
   };



namespace imageplus
{


    namespace kinect_calibration
    {
        int get_intrinsics(Settings& s); //!gets intrinsics parameters
        int get_extrinsics(Settings& s);//!gets extrinsics parameters

        bool runCalibrationAndSave(Settings& s, Size imageSize, Mat&  cameraMatrix, Mat& distCoeffs,
                                   vector<vector<Point2f>> imagePoints );
        bool runCalibrationExtrinsicsAndSave(Settings& s, Size imageSize, Mat  cameraMatrix, Mat distCoeffs,vector<vector<Point2f> > imagePoints );

        void read(const FileNode& node, Settings& x, Settings& default_value);
        double computeReprojectionErrors( const vector<vector<Point3f> >& objectPoints,
                                                 const vector<vector<Point2f> >& imagePoints,
                                                 const vector<Mat>& rvecs, const vector<Mat>& tvecs,
                                                 const Mat& cameraMatrix , const Mat& distCoeffs,
                                                 vector<float>& perViewErrors);
        void calcBoardCornerPositions(Size boardSize, float squareSize, vector<Point3f>& corners,
                                             Settings::Pattern patternType);
        bool runCalibration( Settings& s, Size& imageSize, Mat& cameraMatrix, Mat& distCoeffs,
                                    vector<vector<Point2f> > imagePoints, vector<Mat>& rvecs, vector<Mat>& tvecs,
                                    vector<float>& reprojErrs,  double& totalAvgErr);
        bool runCalibration_extrinsics( Settings& s, Size& imageSize, Mat cameraMatrix, Mat distCoeffs,
                                    vector<vector<Point2f> > imagePoints, Mat& rvecs, Mat& tvecs,
                                    vector<float>& reprojErrs,  double& totalAvgErr);
        void saveCameraParams( Settings& s, Size& imageSize, Mat& cameraMatrix, Mat& distCoeffs,
                                      const vector<Mat>& rvecs, const vector<Mat>& tvecs,
                                      const vector<float>& reprojErrs, const vector<vector<Point2f> >& imagePoints,
                                      double totalAvgErr );
        void saveCameraParams_extrinsics( Settings& s, Size& imageSize, Mat& cameraMatrix, Mat& distCoeffs,
                                      const Mat& rvecs, const Mat& tvecs,
                                      const vector<float>& reprojErrs, const vector<vector<Point2f> >& imagePoints,
                                      double totalAvgErr );
        int getCalibrationParameters (std::string config_file);
        void reproject_2Dto3D();
    }
    namespace kinect_register_reconstruction
    {
        pcl::PointCloud<pcl::PointXYZRGB>::Ptr transform_pointcloud_extrinsics (const pcl::PointCloud<pcl::PointXYZRGB>::ConstPtr &source_cloud,  Eigen::Matrix4d &extrinsics_matrix,bool asus);

        void scene_reconstruction_ownPC(std::string in_file,std::string in_file05,std::string inRGB01_file,std::string inRGB05_file, std::string ex_file,std::string ex_file05,std::string depth01_file, std::string depth05_file,std::string image01_file,std::string image05_file); //reconstrucció des del tauler + ICP


    }
    namespace tools
    {

   void build_pointCloud(Mat& depth, const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud,Mat& cameraMatrix,Mat& distCoeffs,int pc_color);
   void build_pointCloud_RGB(Mat& depth,Mat& image,const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud, Mat& cameraMatrix,Mat& cameraMatrixRGB,Mat& distCoeffs,Mat& distCoeffsRGB,Mat& reg_rotation, Mat& reg_translation);
   //void intersectional_pyramidal_cut (const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud1,const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud2, const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud1_ret,const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud2_ret);
   void intersectional_pyramidal_cut();
   void intersection_pyrmidalCut_extrinsics (const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud05,const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud01_RT, Mat& extr_05, Mat& extr_01 , const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud05_cut,const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud01_RT_cut);

   }
    namespace application
    {
    void fingers_RT(std::string in01_file,std::string in05_file,std::string ex05_file,std::string inRGB01_file,std::string inRGB05_file,vector<std::string> img_list01,vector<std::string> img_list05,vector<Point2f> annoted_fingers01,vector<Point2f> annoted_fingers05,std::string depth_file,std::string depth05_file,std::string image01_file,std::string image05_file,int method);
    void fingers_RT_detector(std::string in_file,std::string in05_file, std::string ex05_file,std::string inRGB01_file,std::string inRGB05_file,std::string detected_fingers01,std::string detected_fingers05,std::string depth_file,std::string depth05_file, std::string image01_file,std::string image05_file);
    }
}


#endif
